#include "glu_kernel_impl.h"
#include "matrix.h"
#include "fillin.h"
#include "symbolic.h"
#include "permutation.h"
#include "numeric.h"

namespace GLU {

KernelImpl::KernelImpl(int n,      /* matrix size */
                       int *ia,    /* row index */
                       int *ja,    /* column index */
                       float *a,   /* matrix A pointer */
                       int nnz,    /* number of non-zeros */
                       float *b,   /* vector b pointer */
                       float *x,   /* unknowns x pointer */
                       std::ostream &out,  /* output stream */
                       std::ostream &err)  /* error stream */
: m_n(n), m_ia(ia), m_ja(ja), m_a(a), m_nnz(nnz), m_b(b), m_x(x), m_out(out),
    m_err(err), m_in(std::cin), m_symbolic(out, err), m_original(out, err),
    m_permutation(0), m_permutated(out, err), m_nthreads(32)
{
    m_out << "Matrix Generated." << std::endl;
}

KernelImpl::KernelImpl(std::istream &in,
                       std::ostream &out,
                       std::ostream &err)
: m_in(in), m_out(out), m_err(err), m_original(out, err),
    m_permutated(out, err), m_symbolic(out, err), m_nthreads(32),
    m_permutation(0)
{
    m_out << "Matrix Generated." << std::endl;
}

// public proxy for operations
void KernelImpl::Do(CMD cmd)
{
    /* proxy */
    switch (cmd) {
    case PARSE:         do_parse();         break;
    case FILLIN:        do_fill_in();       break;
    case PERMUTATION:   do_permutation();   break;
    case SYMBOLIC:      do_symbolic();      break;
    case LEVEL:         do_level();         break;
    case PREDICT:       do_predict();       break;
    case GPUFACT:       do_gpu_fact();      break;
    case SOLVE:         do_solve();         break;
    case CLEAN:         do_clean();         break;
    default:
        m_err << "Unknown command: " << cmd << std::endl;
    }
}

void KernelImpl::do_parse()
{
    m_original.readCooFromStream(m_in);
	m_original.Coo2Csc_Csr();
}

void KernelImpl::do_fill_in()
{
    if (!m_permutation) {
        m_permutation = new int[m_original.row_num];	
    }
	fillin_amd(m_original, m_permutation, m_out, m_err);
}

void KernelImpl::do_permutation()
{
	permutation1(m_original, m_permutation, m_permutated);
}

void KernelImpl::do_symbolic()
{
    m_symbolic.symbolic1(m_permutated);
    m_symbolic.upper_symbolic();
}

void KernelImpl::do_level()
{
    m_symbolic.level();
}

void KernelImpl::do_predict()
{
    m_symbolic.predictLU(m_permutated);
}

void KernelImpl::do_gpu_fact()
{
    LUonDevice(m_symbolic, m_nthreads, m_out, m_err);
}

void KernelImpl::do_solve()
{
    // TODO: solver
}

void KernelImpl::do_clean()
{
    // TODO: free spaces
}

KernelImpl::~KernelImpl()
{}

}; /* namespace GLU */

