#include <cuda.h>
#include <cuda_runtime.h>
#include <iostream>
#include "matrix.h"
#include "symbolic.h"
#include "kernels.h"
#include <cmath>

#define BLOCK_SIZE 1024

//#define THREADNUM 128
#define TMPMEMNUM 300

using namespace std;


void check_error(float *A, float *ref, int dim1, int dim2, bool *abft_error, float Tolerance, int printout)
{
    int error_num = 0;
    for(int col = 0; col < dim2; col++)
    {
        for(int row = 0; row < dim1; row++)
        {
            float error = abs(A[row * dim2 + col] - ref[row * dim2 + col]);
            if (error > Tolerance)
            {
                if (printout)
                {
                    cout << "("<< row << ',' << col << ") , Absolute Difference: " << error <<  endl;
                    cout << "cal: " << A[row * dim2 + col] << "; ref: " << ref[row * dim2 + col] << endl;
                }
                error_num++;
                abft_error[row * dim2 + col] = true;
                // exit(1);         
            }          
        }
    }
    cout << "Error no.: " << error_num << endl;
}

void LUonDevice(Symbolic_Matrix &A_sym, Matrix &A_new, int nthreads, ostream &out, ostream &err, bool perturb, bool printout)
{
	out << "\ncudaLUSolver start." << endl;

	int deviceCount, dev;
	cudaGetDeviceCount(&deviceCount);
	cudaDeviceProp deviceProp;
	dev = 0;
	cudaGetDeviceProperties(&deviceProp, dev);
	cudaSetDevice(dev);
	out << "Device " << dev << ": " << deviceProp.name << " has been selected." << endl;

	cudaEvent_t start, stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	float time;
	
	int n = A_sym.n;
	int nz = A_sym.nnz;
	int num_lev = A_sym.num_lev;

	int *sym_c_ptr_dev, *sym_r_idx_dev;
	float *val_dev;

	int *l_col_ptr_dev, *csr_diag_ptr_dev, *csr_c_idx_dev, *csr_r_ptr_dev;
	int *level_idx_dev, *level_ptr_dev;
	
	cudaMalloc((void**)&sym_c_ptr_dev, (n + 1) * sizeof(int));	
	cudaMalloc((void**)&sym_r_idx_dev, nz * sizeof(int));
	cudaMalloc((void**)&val_dev, nz * sizeof(float));

	cudaMalloc((void**)&l_col_ptr_dev, A_sym.l_col_ptr.size() * sizeof(int));
	cudaMalloc((void**)&csr_diag_ptr_dev, A_sym.csr_diag_ptr.size() * sizeof(int));
	cudaMalloc((void**)&csr_c_idx_dev, A_sym.csr_c_idx.size() * sizeof(int));
	cudaMalloc((void**)&csr_r_ptr_dev, (n + 1) * sizeof(int));

	cudaMalloc((void**)&level_idx_dev, n * sizeof(int));
	cudaMalloc((void**)&level_ptr_dev, A_sym.level_ptr.size() * sizeof(int));

	cudaEventRecord(start, 0);

	cudaMemcpy(sym_c_ptr_dev, &(A_sym.sym_c_ptr[0]), (n + 1) * sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(sym_r_idx_dev, &(A_sym.sym_r_idx[0]), nz * sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(val_dev, &(A_sym.val[0]), nz * sizeof(float), cudaMemcpyHostToDevice);

	cudaMemcpy(l_col_ptr_dev, &(A_sym.l_col_ptr[0]), A_sym.l_col_ptr.size() * sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(csr_diag_ptr_dev, &(A_sym.csr_diag_ptr[0]), A_sym.csr_diag_ptr.size() * sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(csr_c_idx_dev, &(A_sym.csr_c_idx[0]), A_sym.csr_c_idx.size() * sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(csr_r_ptr_dev, &(A_sym.csr_r_ptr[0]), (n + 1) * sizeof(int), cudaMemcpyHostToDevice);

	cudaMemcpy(level_idx_dev, &(A_sym.level_idx[0]), n * sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(level_ptr_dev, &(A_sym.level_ptr[0]), A_sym.level_ptr.size() * sizeof(int), cudaMemcpyHostToDevice);

	float* tmpMem;
	cudaMalloc((void**)&tmpMem, TMPMEMNUM*n*sizeof(float));


// calculate 1-norm of A and perturbation value for perturbation
	float pert = 0;
	if (perturb)
	{
		float norm_A = 0;
		for (int i = 0; i < n; ++i)
		{
			float tmp = 0;
			for (int j = A_sym.sym_c_ptr[i]; j < A_sym.sym_c_ptr[i+1]; ++j)
				tmp += abs(A_sym.val[j]);
			if (norm_A < tmp)
				norm_A = tmp;
		}
		pert = 3.45e-4 * norm_A;
		out << "Gaussian elimination with static pivoting (GESP)..." << endl;
		out << "1-Norm of A matrix is " << norm_A << ", Perturbation value is " << pert << endl;
	}


	// cudaEventRecord(start, 0);
	for (int i = 0; i < num_lev; ++i)
	{
		int lev_size = A_sym.level_size[i];
		dim3 dimBlock(nthreads, 1);
		int j = 0;

		while(lev_size > 0) {
			int restCol = lev_size > TMPMEMNUM ? TMPMEMNUM : lev_size;
			dim3 dimGrid(restCol, 1);

			cudaMemset(tmpMem, 0, TMPMEMNUM*n*sizeof(float));

			if(perturb)
				RL_perturb<<<dimGrid, dimBlock>>>(sym_c_ptr_dev,
									sym_r_idx_dev,
									val_dev,
									l_col_ptr_dev,
									csr_diag_ptr_dev,
									csr_c_idx_dev,
									csr_r_ptr_dev,
									level_idx_dev,
									level_ptr_dev,
									tmpMem,
									n,
									i,
									j*TMPMEMNUM,
									nthreads,
									pert);	
			else
				RL<<<dimGrid, dimBlock>>>(sym_c_ptr_dev,
									sym_r_idx_dev,
									val_dev,
									l_col_ptr_dev,
									csr_diag_ptr_dev,
									csr_c_idx_dev,
									csr_r_ptr_dev,
									level_idx_dev,
									level_ptr_dev,
									tmpMem,
									n,
									i,
									j*TMPMEMNUM,
									nthreads);	
			j++;
			lev_size -= TMPMEMNUM;
		}

	}

	// cudaEventRecord(stop, 0);
	// cudaEventSynchronize(stop);
	// cudaEventElapsedTime(&time, start, stop);


	//copy LU val back to main mem
	cudaMemcpy(&(A_sym.val[0]), val_dev, nz * sizeof(float), cudaMemcpyDeviceToHost);

	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
	cudaEventElapsedTime(&time, start, stop);
	
	out << "cudaLUSolver finished... " << time << " ms" << endl;

	//check NaN elements
	int err_find = 0;
	for(int i = 0; i < nz; i++)
		if(isnan(A_sym.val[i]) || isinf(A_sym.val[i])) 
			err_find++;

	if (err_find == 0)
		cout << "LU data check: No NaN found" << endl;
	else
	{
		cout << "LU data check: " << " NaN found!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
		// exit(-1);
	}

	cudaFree(sym_c_ptr_dev);
	cudaFree(sym_r_idx_dev);
	cudaFree(val_dev);

	cudaFree(l_col_ptr_dev);
	cudaFree(csr_diag_ptr_dev);
	cudaFree(csr_c_idx_dev);
	cudaFree(csr_r_ptr_dev);
	
	cudaFree(level_idx_dev);
	cudaFree(level_ptr_dev);
}

__global__ void dummy(const int* __restrict__ sym_c_ptr_dev,
					  const int* __restrict__ sym_r_idx_dev,
					  float* __restrict__ val_dev,
					  const int* __restrict__ l_col_ptr_dev,
					  const int* __restrict__ csr_diag_ptr_dev,
					  const int* __restrict__ csr_c_idx_dev,
					  const int* __restrict__ csr_r_ptr_dev,
					  const int* __restrict__ level_idx_dev,
					  const int* __restrict__ level_ptr_dev,
					  const int level,
                      const int nthreads)
{
	const int tid = threadIdx.x;
	const int bid = blockIdx.x;

	const int currentCol = level_idx_dev[level_ptr_dev[level]+bid];
	const int currentColSize = sym_c_ptr_dev[currentCol+1] - sym_c_ptr_dev[currentCol];
	const int currentPos = sym_c_ptr_dev[currentCol] + tid; 
	
	if (tid < currentColSize)
	{
		val_dev[currentPos] = 1;	
	}
}


__global__ void RL_LU(const int* __restrict__ sym_c_ptr_dev,
					  const int* __restrict__ sym_r_idx_dev,
					  float* __restrict__ val_dev,
					  const int* __restrict__ l_col_ptr_dev,
					  const int* __restrict__ csr_diag_ptr_dev,
					  const int* __restrict__ csr_c_idx_dev,
					  const int* __restrict__ csr_r_ptr_dev,
					  const int* __restrict__ level_idx_dev,
					  const int* __restrict__ level_ptr_dev,
					  const int level,
                      const int nthreads)
{
	const int tid = threadIdx.x;
	const int bid = blockIdx.x;
	const int wid = threadIdx.x / 32;
	
	const int currentCol = level_idx_dev[level_ptr_dev[level]+bid];
	const int currentLColSize = sym_c_ptr_dev[currentCol + 1] - l_col_ptr_dev[currentCol] - 1;
	const int currentLPos = l_col_ptr_dev[currentCol] + tid + 1; 
	//update current col
	int offset = 0;
	while (currentLColSize > offset)
	{
		if (currentLPos + offset < currentLColSize)
		{
			val_dev[currentLPos + offset] /= val_dev[l_col_ptr_dev[currentCol]];
		}
		offset += nthreads;
	}
	__syncthreads();

	//broadcast to submatrix
	const int subColPos = csr_diag_ptr_dev[currentCol] + wid + 1;
	const int subMatSize = csr_r_ptr_dev[currentCol + 1] - csr_diag_ptr_dev[currentCol] - 1;
	const int subCol = csr_c_idx_dev[subColPos];
	const int tidInWarp = threadIdx.x % 32;
	const int subColElem = sym_c_ptr_dev[subCol] + tidInWarp; 
	const int currentLElem = l_col_ptr_dev[currentCol] + tidInWarp + 1;

	int woffset = 0;
	while (subMatSize > woffset)
	{
		if (wid + woffset < subMatSize)		
		{
			while (currentLColSize > offset)
			{
				if (tidInWarp + offset < currentLColSize)
				{	
					val_dev[subColElem + offset] -= val_dev[currentLElem]*1.5;
				}
				offset += nthreads;
			}
		}
		woffset += nthreads/32;
	}
}


__global__ void LL_LU(const int* __restrict__ sym_c_ptr_dev,
					  const int* __restrict__ sym_r_idx_dev,
					  float* __restrict__ val_dev,
					  const int* __restrict__ l_col_ptr_dev,
					  const int* __restrict__ csr_diag_ptr_dev,
					  const int* __restrict__ level_idx_dev,
					  const int* __restrict__ level_ptr_dev,
					  const int level,
                      const int nthreads)
{
	const int tid = threadIdx.x;
	const int bid = blockIdx.x;

	const int currentCol = level_idx_dev[level_ptr_dev[level]+bid];
	const int currentLColSize = sym_c_ptr_dev[currentCol + 1] - l_col_ptr_dev[currentCol] - 1;
	const int currentLPos = l_col_ptr_dev[currentCol] + tid + 1; 
	int offset = 0;

	for (int i = sym_c_ptr_dev[currentCol]; i < l_col_ptr_dev[currentCol]; ++i)
	{
		float tmpVal = val_dev[i];
		int updateCol = sym_r_idx_dev[i];
		int updateColSize = sym_c_ptr_dev[updateCol+1] - l_col_ptr_dev[updateCol] - 1;
		
		while (updateColSize > offset)
		{
			if (tid + offset < updateColSize)
			{
				val_dev[i + tid + offset + 1] -= tmpVal * val_dev[l_col_ptr_dev[updateCol] + tid + offset + 1];
			}
			offset += nthreads;
		}
		__syncthreads();
	}

	offset = 0;
	__syncthreads();
	
	while (currentLColSize > offset)
	{
		if (currentLPos + offset < currentLColSize)
		{
			val_dev[currentLPos + offset] /= val_dev[l_col_ptr_dev[currentCol]];
		}
		offset += nthreads;
	}
}					  


__global__ void LL_LU_new(const int* __restrict__ sym_c_ptr_dev,
					      const int* __restrict__ sym_r_idx_dev,
						  float* __restrict__ val_dev,
					      const int* __restrict__ l_col_ptr_dev,
					      const int* __restrict__ csr_diag_ptr_dev,
					      const int* __restrict__ level_idx_dev,
					      const int* __restrict__ level_ptr_dev,
					      const int level,
                          const int nthreads)
{
	const int tid = threadIdx.x;
	const int bid = blockIdx.x;

	const int currentCol = level_idx_dev[level_ptr_dev[level]+bid];
	const int currentLColSize = sym_c_ptr_dev[currentCol + 1] - l_col_ptr_dev[currentCol] - 1;
	const int currentLPos = l_col_ptr_dev[currentCol] + tid + 1; 
	int offset = 0;

	for (int i = sym_c_ptr_dev[currentCol]; i < l_col_ptr_dev[currentCol]; ++i)
	{
		float tmpVal = val_dev[i];
		int updateCol = sym_r_idx_dev[i];
		int updateColSize = sym_c_ptr_dev[updateCol+1] - l_col_ptr_dev[updateCol] - 1;
		int updateColPos = tid + l_col_ptr_dev[updateCol] + 1;

		while (updateColSize > offset)
		{
			if (tid + offset < updateColSize)
			{
				float updateVal = tmpVal * val_dev[updateColPos + offset];
				int updateRidx = sym_r_idx_dev[updateColPos + offset];
				int tmpPos = i + tid + offset + 1;
				while (updateRidx > sym_r_idx_dev[tmpPos] && tmpPos < sym_c_ptr_dev[currentCol+1])
				{
					tmpPos++;
				}
				if (tmpPos < sym_c_ptr_dev[currentCol+1])
				{
					val_dev[tmpPos] -= updateVal;
				}
			}
			offset += nthreads;
		}
		__syncthreads();
	}

	offset = 0;
	__syncthreads();
	
	while (currentLColSize > offset)
	{
		if (currentLPos + offset < currentLColSize)
		{
			val_dev[currentLPos + offset] /= val_dev[l_col_ptr_dev[currentCol]];
		}
		offset += nthreads;
	}
}					  

__global__ void RL_LU_new(const int* __restrict__ sym_c_ptr_dev,
					      const int* __restrict__ sym_r_idx_dev,
					      float* __restrict__ val_dev,
					      const int* __restrict__ l_col_ptr_dev,
					      const int* __restrict__ csr_diag_ptr_dev,
					      const int* __restrict__ csr_c_idx_dev,
					      const int* __restrict__ csr_r_ptr_dev,
					      const int* __restrict__ level_idx_dev,
					      const int* __restrict__ level_ptr_dev,
					      const int level,
                          const int nthreads)
{
	const int tid = threadIdx.x;
	const int bid = blockIdx.x;
	const int wid = threadIdx.x / 32;
	
	const int currentCol = level_idx_dev[level_ptr_dev[level]+bid];
	const int currentLColSize = sym_c_ptr_dev[currentCol + 1] - l_col_ptr_dev[currentCol] - 1;
	const int currentLPos = l_col_ptr_dev[currentCol] + tid + 1; 
	//update current col
	int offset = 0;
	while (currentLColSize > offset)
	{
		if (currentLPos + offset < currentLColSize)
		{
			val_dev[currentLPos + offset] /= val_dev[l_col_ptr_dev[currentCol]];
		}
		offset += nthreads;
	}
	__syncthreads();

	offset = 0;
	//broadcast to submatrix
	const int subColPos = csr_diag_ptr_dev[currentCol] + wid + 1;
	const int subMatSize = csr_r_ptr_dev[currentCol + 1] - csr_diag_ptr_dev[currentCol] - 1;
	int subCol = csr_c_idx_dev[subColPos];
	const int tidInWarp = threadIdx.x % 32;
	int subColElem = sym_c_ptr_dev[subCol] + tidInWarp + 1; 
	int currentLElem = l_col_ptr_dev[currentCol] + tidInWarp + 1;

	int woffset = 0;
	while (subMatSize > woffset)
	{
		if (wid + woffset < subMatSize)		
		{
			subCol = csr_c_idx_dev[subColPos + woffset];
			while (currentLColSize > offset)
			{
				if (tidInWarp + offset < currentLColSize)
				{
					int updateRidx = sym_r_idx_dev[currentLElem + offset];	
					while (updateRidx > sym_r_idx_dev[subColElem + offset] && subColElem + offset< sym_c_ptr_dev[subCol+1])
					{
						subColElem++;
					}
					if (subColElem +offset < sym_c_ptr_dev[subCol+1])
					{
						val_dev[subColElem + offset] -= val_dev[currentLElem + offset];
					}
				}
				offset += nthreads;
			}
		}
		woffset += nthreads/32;
	}
}

__global__ void RL(const int* __restrict__ sym_c_ptr_dev,
					      const int* __restrict__ sym_r_idx_dev,
					      float* __restrict__ val_dev,
					      const int* __restrict__ l_col_ptr_dev,
					      const int* __restrict__ csr_diag_ptr_dev,
					      const int* __restrict__ csr_c_idx_dev,
					      const int* __restrict__ csr_r_ptr_dev,
					      const int* __restrict__ level_idx_dev,
					      const int* __restrict__ level_ptr_dev,
					      float* __restrict__ tmpMem,
					      const int n,
					      const int level,
					      const int inLevPos,
						  const int nthreads)
{
	const int tid = threadIdx.x;
	const int bid = blockIdx.x;
	const int wid = threadIdx.x / 32;

	const int currentCol = level_idx_dev[level_ptr_dev[level]+inLevPos+bid];
	const int currentLColSize = sym_c_ptr_dev[currentCol + 1] - l_col_ptr_dev[currentCol] - 1;
	const int currentLPos = l_col_ptr_dev[currentCol] + tid + 1;

	__shared__ float s[32];

	//update current col

	int offset = 0;
	while (currentLColSize > offset)
	{
		if (tid + offset < currentLColSize)
		{
			int ridx = sym_r_idx_dev[currentLPos + offset];
			
			val_dev[currentLPos + offset] /= val_dev[l_col_ptr_dev[currentCol]];
			tmpMem[bid*n + ridx]= val_dev[currentLPos + offset];
		}
		offset += nthreads;
	}
	__syncthreads();

	offset = 0;
	//broadcast to submatrix
	const int subColPos = csr_diag_ptr_dev[currentCol] + wid + 1;
	const int subMatSize = csr_r_ptr_dev[currentCol + 1] - csr_diag_ptr_dev[currentCol] - 1;
	int subCol = csr_c_idx_dev[subColPos];
	const int tidInWarp = threadIdx.x % 32;
	int subColElem = 0;

	int woffset = 0;
	while (subMatSize > woffset)
	{
		if (wid + woffset < subMatSize)
		{
			offset = 0;
			subCol = csr_c_idx_dev[subColPos + woffset];
			while(offset < sym_c_ptr_dev[subCol + 1] - sym_c_ptr_dev[subCol])
			{
				if (tidInWarp + offset < sym_c_ptr_dev[subCol + 1] - sym_c_ptr_dev[subCol])
				{

					subColElem = sym_c_ptr_dev[subCol] + tidInWarp + offset;
					int ridx = sym_r_idx_dev[subColElem];

					if (ridx == currentCol) 
					{
						s[wid] = val_dev[subColElem];
					}
					__syncthreads();
					if (ridx > currentCol) 
					{
						atomicAdd(&val_dev[subColElem], -tmpMem[ridx+n*bid]*s[wid]);
					}
				}
				offset += 32;
			}
		}
		woffset += nthreads/32;
	}

}

__global__ void RL_perturb(const int* __restrict__ sym_c_ptr_dev,
					      const int* __restrict__ sym_r_idx_dev,
					      float* __restrict__ val_dev,
					      const int* __restrict__ l_col_ptr_dev,
					      const int* __restrict__ csr_diag_ptr_dev,
					      const int* __restrict__ csr_c_idx_dev,
					      const int* __restrict__ csr_r_ptr_dev,
					      const int* __restrict__ level_idx_dev,
					      const int* __restrict__ level_ptr_dev,
					      float* __restrict__ tmpMem,
					      const int n,
					      const int level,
					      const int inLevPos,
						  const int nthreads,
						  const float pert)
{
	const int tid = threadIdx.x;
	const int bid = blockIdx.x;
	const int wid = threadIdx.x / 32;

	const int currentCol = level_idx_dev[level_ptr_dev[level]+inLevPos+bid];
	const int currentLColSize = sym_c_ptr_dev[currentCol + 1] - l_col_ptr_dev[currentCol] - 1;
	const int currentLPos = l_col_ptr_dev[currentCol] + tid + 1;

	__shared__ float s[32];

	//update current col

	int offset = 0;
	while (currentLColSize > offset)
	{
		if (tid + offset < currentLColSize)
		{
			int ridx = sym_r_idx_dev[currentLPos + offset];

			if (abs(val_dev[l_col_ptr_dev[currentCol]]) < pert)
				val_dev[l_col_ptr_dev[currentCol]] = pert;
			
			val_dev[currentLPos + offset] /= val_dev[l_col_ptr_dev[currentCol]];
			tmpMem[bid*n + ridx]= val_dev[currentLPos + offset];
		}
		offset += nthreads;
	}
	__syncthreads();

	offset = 0;
	//broadcast to submatrix
	const int subColPos = csr_diag_ptr_dev[currentCol] + wid + 1;
	const int subMatSize = csr_r_ptr_dev[currentCol + 1] - csr_diag_ptr_dev[currentCol] - 1;
	int subCol = csr_c_idx_dev[subColPos];
	const int tidInWarp = threadIdx.x % 32;
	int subColElem = 0;

	int woffset = 0;
	while (subMatSize > woffset)
	{
		if (wid + woffset < subMatSize)
		{
			offset = 0;
			subCol = csr_c_idx_dev[subColPos + woffset];
			while(offset < sym_c_ptr_dev[subCol + 1] - sym_c_ptr_dev[subCol])
			{
				if (tidInWarp + offset < sym_c_ptr_dev[subCol + 1] - sym_c_ptr_dev[subCol])
				{

					subColElem = sym_c_ptr_dev[subCol] + tidInWarp + offset;
					int ridx = sym_r_idx_dev[subColElem];

					if (ridx == currentCol) 
					{
						s[wid] = val_dev[subColElem];
					}
					__syncthreads();
					if (ridx > currentCol) 
					{
						atomicAdd(&val_dev[subColElem], -tmpMem[ridx+n*bid]*s[wid]);
					}
				}
				offset += 32;
			}
		}
		woffset += nthreads/32;
	}

}
